/*
 * IP_MASQ_RTSP - RTSP masquerading module for Linux
 * Author:  Tom Marshall, RealNetworks Inc. <tmarshall@real.com>
 * Version: 0.91
 *
 *	 This program is free software; you can redistribute it and/or
 *	 modify it under the terms of the GNU General Public License
 *	 as published by the Free Software Foundation; either version
 *	 2 of the License, or (at your option) any later version.
 *
 * Changes:
 *   0.91: masq_rtsp_init: terminate for() loop after the first 0 element
 *         in ports[] is encountered.
 *   0.90: initial version.
 *
 * Features, limitations, and random notes:
 *
 *   - This is a complete rewrite of the code.  Emphasis has been placed
 *     on correctness.  It has been written and tested on the 2.2.18 kernel
 *     but should work with any recent 2.2.x kernel.  I have heard rumor that
 *     the masq code changed slightly somewhere around 2.2.5 so you will
 *     probably want 2.2.6 or later.  If you know exactly what changed and
 *     when, I would be interested in knowing.
 *
 *   - This module doesn't keep any state between packets, so there is a
 *     (extremely remote) possibility that we will misinterpret the data.
 *
 *   - In order to be "seen", an RTSP message must be contained within a
 *     single packet, or at least begin on a packet boundary and include the
 *     transport and cseq headers.  It is conceivable that messages could
 *     span packets (suppose the headers are written one line at a time).
 *     I am not aware of any apps that do such a silly thing.
 *
 *   - Multiple RTSP messages in a single packet are handled, including
 *     entities (note: this is mostly untested).  TCP interleaved data is
 *     not handled, but support could/should be added.
 *
 *   - Multicast transports are ignored (we should probably strip them).
 *
 *   - Encoding/recording (eg. RealProducer, SLTA) is fully supported,
 *     provided that the server's encoder port is in the list.  SLTA worked
 *     fine in my tests, but RealProducer seemed to be a bit shaky (lots of
 *     rebuffering on the clients).  I don't think that was related to this
 *     module, but I can't say for sure.
 *
 *   - client_port mapping uses the cseq to track request/response pairs
 *     because we cannot rely on the client_port values in the server's
 *     response (ie, if another firewall running a broken masq module is
 *     between us and the server).  In an ideal world, we would just lookup
 *     the returned client_port value in a mapping table and replace it.
 *
 *   - We don't track teardowns to release the mports.  There are two reasons
 *     for this: (1) the masq code relies on timeouts, there is no mport
 *     delete function, and (2) the RTSP protocol makes it prohibitive.  We
 *     would have to track sessions, which are tokens of arbitrary size.  It
 *     would be nice if the client_port was sent in the teardown request.
 *
 *   - RTP port allocation sucks.  It's an irritant on the client side, it's
 *     a pain on the server side, and it's downright hostile in between.
 *     There is no reason that RTP couldn't have been designed to use two
 *     unrelated UDP ports, or even a single UDP port.  Whoever designs
 *     standard internet protocols should be required to either have done
 *     proxy/firewall coding or consult with someone who was.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/init.h>
#include <net/protocol.h>
#include <net/tcp.h>

#include <net/ip_masq.h>

/*** BEGIN TUNABLE PARAMETERS ***/

/* RTP port allocation strategy (see alloc_ports) */
#define RTP_ALLOC_STRATEGY 3

/* Max simultaneous (pipelined) SETUP requests to track */
#define MAX_SIMUL_SETUP 4

/*
 * Max mappings we will allow for one RTSP connection (for RTP, the number
 * of allocated ports is twice this value).  Note that SMIL burns a lot of
 * ports so keep this reasonably high.  If this is too low, you will see a
 * lot of "no free client map entries" messages.
 */
#define MAX_PORT_MAPS 16

/* The default port list */
static int ports[MAX_MASQ_APP_PORTS] =
{
    554,    /* Standard RTSP */
    3030,   /* RealNetworks pull splitter */
    4040,   /* RealNetworks live encoder */
    0
}; 
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_MASQ_APP_PORTS) "i");

/*** END TUNABLE PARAMETERS ***/

#define IS_DIGIT(c) (c >= '0' && c <= '9')
#define IS_WSPACE(c) (c == ' ' || c == '\t' || c == '\r' || c == '\n')
#define SKIP_WSPACE(p,n) while( n > 0 && IS_WSPACE(*p) ) { n--; p++; }

/* return flags for parse_rtsp_message() */
#define RTSPF_MATCH (1<<0)
#define RTSPF_TRAN  (1<<1)
#define RTSPF_CSEQ  (1<<2)
#define RTSPF_ALL   ((1<<3)-1)

/*
 * If sport_lo is zero, the map entry is unused.  The entries are used in
 * order and remain active for the duration of the connection.
 *
 * If cseq[n] is zero, it is unused.  They should be allocated in order and
 * deallocated in reverse order, but we don't rely on that.
 *
 * All port values are host endian.
 */
struct rtsp_port_map
{
    __u16           sport_lo;
    __u16           sport_hi;
    __u16           mport_lo;
    __u16           mport_hi;
    __u32           cseq[MAX_SIMUL_SETUP];
};
#define APP_DATA_LEN (sizeof(struct rtsp_port_map)*MAX_PORT_MAPS)

static struct ip_masq_app* g_masq_objs[MAX_MASQ_APP_PORTS];

/*** helpers ***/

static int
strncasecmp( char* s1, char* s2, __u32 n )
{
    while( n-- )
    {
        if( ((*s1)|0x20) != ((*s2)|0x20) )
        {
            return *s1 - *s2;
        }
        s1++;
        s2++;
    }
    return 0;
}

static int
str_to_u16( char* pbuf, __u16* pval )
{
    int n = 0;

    *pval = 0;
    while( pbuf[n] >= '0' && pbuf[n] <= '9' )
    {
        *pval = (*pval * 10) + (pbuf[n] - '0');
        n++;
    }

    return n;
}

static int
str_to_u32( char* pbuf, __u32* pval )
{
    int n = 0;

    *pval = 0;
    while( pbuf[n] >= '0' && pbuf[n] <= '9' )
    {
        *pval = (*pval * 10) + (pbuf[n] - '0');
        n++;
    }

    return n;
}

static int
next_line( char** ppbuf, __u32* pbuflen, char** ppline, __u32* plinelen )
{
    char* pbuf = *ppbuf;
    __u32 buflen = *pbuflen;
    __u32 linelen = 0;

    do
    {
        while( *pbuf != '\r' && *pbuf != '\n' )
        {
            if( buflen <= 1 )
            {
                return 0;
            }

            pbuf++;
            linelen++;
            buflen--;
        }

        if( buflen > 1 && *pbuf == '\r' && *(pbuf+1) == '\n' )
        {
            pbuf++;
            buflen--;
        }

        pbuf++;
        buflen--;
    }
    while( buflen > 0 && (*pbuf == ' ' || *pbuf == '\t') );

    *ppline = *ppbuf;
    *plinelen = linelen;
    *ppbuf = pbuf;
    *pbuflen = buflen;

    return linelen;
}

static void
get_skb_data( struct sk_buff** ppskb, char** ppdata, __u32* plen )
{
    struct sk_buff* skb;
    struct iphdr*   iph;
    struct tcphdr*  th;
    char* pbuf;
    char* pend;

    skb = *ppskb;
    iph = skb->nh.iph;
    th = (struct tcphdr*)&(((char*)iph)[iph->ihl*4]);

    pbuf = (char*)th + (th->doff * 4);
    pend = skb->h.raw + skb->len;
    *ppdata = pbuf;
    *plen = pend - pbuf;
}

/*
 * Allocate a UDP port (pair)
 *
 * This gets tricky because:
 *   - ip_masq_new() will not necessarily return sequential ports.
 *   - there isn't a way to release a port manually so repeated ip_masq_new()
 *     calls burn up ports quickly in a nondeterministic manner (but they
 *     should be freed quickly).
 *   - there isn't a way to lock the ip_masq structures while we search so
 *     using ip_masq_in_get() to check for free ports has a race condition,
 *     in addition to relying on the PORT_MASQ_BEGIN/PORT_MASQ_END macros
 *     matching the running kernel.
 *
 * The best we can do is to have a race condition in here, unless we make
 * some modification to the ip_masq kernel interface.  I'm not sure if there
 * is actually a race condition here with a single cpu, but you can bet that
 * there's one for SMP.
 *
 * The source from Apple doesn't burn ports but it doesn't do adequate checks
 * on the ports either.  This has the potential to both confuse the server and
 * drop packets.
 */
static int
alloc_ports( __u32 maddr, struct ip_masq* pms, __u16 sport_lo, __u16 sport_hi, struct rtsp_port_map* preq )
{
#if RTP_ALLOC_STRATEGY == 2
    __u16 port;
#endif
    struct ip_masq* pms_new[2];
    pms_new[0] = pms_new[1] = NULL;

    if( sport_lo == sport_hi ) /* Unlikely but possible */
    {
        pms_new[0] = ip_masq_new( IPPROTO_UDP, maddr, 0,
                                  pms->saddr, sport_lo,
                                  pms->daddr, 0,
                                  IP_MASQ_F_NO_DPORT );
        if( ! pms_new[0] )
        {
            return 0;
        }

        ip_masq_listen( pms_new[0] );
        ip_masq_control_add( pms_new[0], pms );
        ip_masq_put( pms_new[0] );

        preq->sport_lo = preq->sport_hi = ntohs(sport_lo);
        preq->mport_lo = preq->mport_hi = ntohs(pms_new[0]->mport);

        return 1;
    }

#if RTP_ALLOC_STRATEGY == 1
    /* Strategy 1: Kernel masq support for port blocks - not yet written */
    ip_masq_new_block( pms_new, 2, IPPROTO_UDP, maddr, 0,
                       pms->saddr, sport_lo,
                       pms->daddr, 0,
                       IP_MASQ_F_NO_DPORT );
    if( ! pms_new[0] || ! pms_new[1] )
    {
        return 0;
    }
    ip_masq_listen( pms_new[0] );
    ip_masq_control_add( pms_new[0], pms );
    ip_masq_put( pms_new[0] );
    ip_masq_listen( pms_new[1] );
    ip_masq_control_add( pms_new[1], pms );
    ip_masq_put( pms_new[1] );
#elif RTP_ALLOC_STRATEGY == 2
    /* Strategy 2: Manual search with race condition */
    for( port = PORT_MASQ_BEGIN; port < PORT_MASQ_END; port += 2 )
    {
        if( ! ip_masq_in_get( IPPROTO_UDP, pms->saddr, sport_lo, maddr, htons(port) ) &&
            ! ip_masq_in_get( IPPROTO_UDP, pms->saddr, sport_hi, maddr, htons(port+1) ) )
        {
            pms_new[0] = ip_masq_new( IPPROTO_UDP, maddr, htons(port),
                                      pms->saddr, sport_lo,
                                      pms->daddr, 0,
                                      IP_MASQ_F_NO_DPORT );
            if( ! pms_new[0] )
            {
                return 0;
            }
            ip_masq_listen( pms_new[0] );
            ip_masq_control_add( pms_new[0], pms );
            ip_masq_put( pms_new[0] );
            pms_new[1] = ip_masq_new( IPPROTO_UDP, maddr, htons(port+1),
                                      pms->saddr, sport_hi,
                                      pms->daddr, 0,
                                      IP_MASQ_F_NO_DPORT );
            if( ! pms_new[1] )
            {
                return 0;
            }
            ip_masq_listen( pms_new[1] );
            ip_masq_control_add( pms_new[1], pms );
            ip_masq_put( pms_new[1] );

            break;
        }
    }
#elif RTP_ALLOC_STRATEGY == 3
    /* Strategy 3: Burn ports with race condition */
    do
    {
        do
        {
            pms_new[0] = ip_masq_new( IPPROTO_UDP, maddr, 0,
                                      pms->saddr, sport_lo,
                                      pms->daddr, 0,
                                      IP_MASQ_F_NO_DPORT );
            if( ! pms_new[0] )
            {
                return 0;
            }
            ip_masq_listen( pms_new[0] );
            ip_masq_control_add( pms_new[0], pms );
            ip_masq_put( pms_new[0] );
        }
        while( ntohs(pms_new[0]->mport) & 1 );

        pms_new[1] = ip_masq_new( IPPROTO_UDP, maddr, 0,
                                  pms->saddr, sport_hi,
                                  pms->daddr, 0,
                                  IP_MASQ_F_NO_DPORT );
        if( ! pms_new[1] )
        {
            return 0;
        }
        ip_masq_listen( pms_new[1] );
        ip_masq_control_add( pms_new[1], pms );
        ip_masq_put( pms_new[1] );
    }
    while( ntohs(pms_new[1]->mport) != ntohs(pms_new[0]->mport)+1 );
#else
#error "I need an RTP_ALLOC_STRATEGY"
#endif /* RTP_ALLOC_STRATEGY */

    preq->sport_lo = ntohs(sport_lo);
    preq->sport_hi = ntohs(sport_hi);
    preq->mport_lo = ntohs(pms_new[0]->mport);
    preq->mport_hi = ntohs(pms_new[1]->mport);

    return 1;
}

/*
 * Parse an RTSP packet.
 * Returns zero if parsing failed, else a bitfield:
 *   RTSPF_MATCH: the message started with pmatch
 *   RTSPF_TRAN:  there is a transport header
 *   RTSPF_CSEQ:  there is a cseq header
 *
 * In/Out parameters:
 *   *ppbuf is current buffer ptr
 *   *pbuflen is current buffer len
 * Out parameters:
 *   *pcseq set to CSeq, if found
 *   *pptran set to Transport, if found
 */
static __u32
parse_rtsp_message( char** ppbuf, __u32* pbuflen, char* pmatch, __u32* pcseq, char** pptran, __u32* ptranlen )
{
    int   rv;
    __u32 matchlen;
    __u32 entitylen;
    char* pline;
    __u32 linelen;

    rv = 0;
    matchlen = strlen( pmatch );
    entitylen = 0;

    if( !next_line( ppbuf, pbuflen, &pline, &linelen ) )
    {
        return 0;
    }
    if( strncasecmp( pline, pmatch, matchlen ) == 0 )
    {
        rv |= RTSPF_MATCH;
    }

    while( next_line( ppbuf, pbuflen, &pline, &linelen ) )
    {
        if( linelen == 0 )
        {
            break;
        }
        if( linelen > 10 && strncasecmp( pline, "Transport:", 10 ) == 0 )
        {
            *pptran = pline;
            *ptranlen = linelen;
            rv |= RTSPF_TRAN;
        }
        else if( linelen > 5 && strncasecmp( pline, "CSeq:", 5 ) == 0 )
        {
            pline += 5;
            SKIP_WSPACE( pline, linelen );
            str_to_u32( pline, pcseq );
            rv |= RTSPF_CSEQ;
        }
        else if( linelen > 15 && strncasecmp( pline, "Content-Length:", 15 ) == 0 )
        {
            pline += 15;
            SKIP_WSPACE( pline, linelen );
            str_to_u32( pline, &entitylen );
        }
    }
    if( entitylen > 0 && linelen == 0 )
    {
        if( entitylen > *pbuflen )
        {
            entitylen = *pbuflen;
        }
        *ppbuf += entitylen;
        *pbuflen -= entitylen;
    }

    return rv;
}

/*
 * Find lo/hi client ports (if any) in transport header
 * In:
 *   ptran, tranlen = buffer to search
 *
 * Out:
 *   pport_lo, pport_hi = lo/hi ports (host endian)
 *
 * Returns nonzero if any client ports found
 *
 * Note: it is valid (and expected) for the client to request multiple
 * transports, so we need to parse the entire line.
 */
static int
get_transport_ports( char* ptran, __u32 tranlen, __u16* pport_lo, __u16* pport_hi )
{
    int rv = 0;
    __u32 numlen;

    if( strncasecmp( ptran, "Transport:", 10 ) != 0 )
    {
        printk( KERN_INFO "ip_masq_rtsp: get_transport_ports: sanity check failed\n" );
        return 0;
    }

    *pport_lo = *pport_hi = 0;
    ptran += 10;
    SKIP_WSPACE( ptran, tranlen );

    while( tranlen > 0 )
    {
        if( *ptran == ';' && strncmp( ptran, ";client_port=", 13 ) == 0 )
        {
            ptran += 13; tranlen -= 13;
            numlen = str_to_u16( ptran, pport_lo );
            ptran += numlen; tranlen -= numlen;
            *pport_hi = *pport_lo;
            if( *ptran == '-' )
            {
                ptran++; tranlen--;
                numlen = str_to_u16( ptran, pport_hi );
                ptran += numlen; tranlen -= numlen;
            }
            rv = 1;
        }
        else
        {
            ptran++; tranlen--;
        }
    }

    return rv;
}

/*
 * Replace all occurences of "client_port=<spec>"
 * In:
 *   ppskb      = packet
 *   ptran      = location of "Transport:" header
 *   tranlen    = length of "Transport:" header
 *   rport_lo   = replacement low port (host endian)
 *   rport_hi   = replacement high port (host endian)
 *
 * Returns packet size difference.
 *
 * Assumes that a complete transport header is present, ending with CR or LF
 */
static int
replace_client_port( struct sk_buff** ppskb, char* ptran, __u32 tranlen, __u16 rport_lo, __u16 rport_hi )
{
    int   pos;
    int   diff;
    char* skb_data;
    int   skb_len;
    char* obuf;             /* Original buffer */
    int   obuf_len;         /* Original len */
    char  rbuf1[16];        /* Replacement buffer (one port) */
    int   rbuf1_len = 0;    /* Replacement len (one port) */
    char  rbufa[16];        /* Replacement buffer (all ports) */
    int   rbufa_len = 0;    /* Replacement len (all ports) */

    diff = 0;
    get_skb_data( ppskb, &skb_data, &skb_len );
    pos = ptran - skb_data;

    if( tranlen < 10 || pos >= skb_len-10 ||
        strncasecmp( &skb_data[pos], "Transport:", 10 ) != 0 )
    {
        printk( KERN_INFO "ip_masq_rtsp: replace_client_port: sanity check failed\n" );
        return 0;
    }

    pos += 10; tranlen -= 10;
    while( tranlen > 0 && IS_WSPACE(skb_data[pos]) ) { pos++; tranlen--; }

    rbuf1_len = sprintf( rbuf1, "%hu", rport_lo );
    rbufa_len = sprintf( rbufa, "%hu-%hu", rport_lo, rport_hi );

    while( tranlen > 0 )
    {
        if( pos >= skb_len )
        {
            printk( KERN_INFO "ip_masq_rtsp: replace_client_port: pos overrun\n" );
            return diff;
        }

        /* tran;param;param,tran;param;param,... */
        if( skb_data[pos] == ';' && strncmp( &skb_data[pos], ";client_port=", 13 ) == 0 )
        {
            char* rbuf = rbuf1;
            int rbuf_len = rbuf1_len;

            pos += 13; tranlen -= 13;
            obuf = &skb_data[pos];
            obuf_len = 0;
            while( IS_DIGIT(skb_data[pos]) || skb_data[pos] == '-' )
            {
                if( skb_data[pos] == '-' )
                {
                    rbuf = rbufa;
                    rbuf_len = rbufa_len;
                }
                obuf_len++;
                pos++; tranlen--;

                if( pos >= skb_len )
                {
                    printk( KERN_INFO "ip_masq_rtsp: replace_client_port: pos overrun\n" );
                    return diff;
                }
            }

            if( obuf_len == rbuf_len ) /* Unlikely but possible */
            {
                memcpy( obuf, rbuf, rbuf_len );
            }
            else
            {
                *ppskb = ip_masq_skb_replace( *ppskb, GFP_ATOMIC, obuf, obuf_len, rbuf, rbuf_len );
                get_skb_data( ppskb, &skb_data, &skb_len );
                diff += rbuf_len - obuf_len;
            }
        }
        else
        {
            pos++; tranlen--;
        }
    }

    return diff;
}

/**************************************
 * masq funcs
 **************************************/

static int
masq_rtsp_init_1( struct ip_masq_app* pmapp, struct ip_masq* pms )
{
    MOD_INC_USE_COUNT;

    pms->app_data = kmalloc( APP_DATA_LEN, GFP_ATOMIC );
    if( pms->app_data == NULL )
    {
        printk( KERN_INFO "ip_masq_rtsp: no memory for application data\n" );
        return 0;
    }

    memset( pms->app_data, 0, APP_DATA_LEN );

    return 0;
}

static int
masq_rtsp_done_1( struct ip_masq_app* pmapp, struct ip_masq* pms )
{
    MOD_DEC_USE_COUNT;

    if( pms->app_data )
    {
        kfree_s( pms->app_data, APP_DATA_LEN );
    }

    return 0;
}

/*
 * Handle incoming (S->C) packet
 */
static int
masq_rtsp_in( struct ip_masq_app* papp, struct ip_masq* pms, struct sk_buff** ppskb, __u32 maddr )
{
    int   diff     = 0;
    __u32 skblen   = 0;
    char* pdata = NULL;
    __u32 datalen  = 0;
    __u32 cseq     = 0;
    char* ptran    = NULL;
    __u32 tranlen  = 0;

    int i = 0;
    int j = 0;
    struct rtsp_port_map* pmap;
    struct rtsp_port_map* preq;
    __u32 rtspf;

    get_skb_data( ppskb, &pdata, &skblen );
    datalen = skblen;

    while( datalen )
    {
        pmap = (struct rtsp_port_map*)pms->app_data;
        rtspf = parse_rtsp_message( &pdata, &datalen, "RTSP/", &cseq, &ptran, &tranlen );

        /* Is it a response with transport and cseq? */
        if( rtspf != RTSPF_ALL )
        {
            if( rtspf == 0 )
            {
                break; /* Parse failed, abort */
            }
            continue; /* Parse succeeded, keep going */
        }

        /* Is it a setup response? */
        preq = NULL;
        for( i = 0; preq == NULL && i < MAX_PORT_MAPS; i++ )
        {
            if( pmap[i].sport_lo == 0 )
            {
                continue;
            }

            for( j = 0; j < MAX_SIMUL_SETUP; j++ )
            {
                if( pmap[i].cseq[j] == cseq )
                {
                    preq = &pmap[i];
                    break;
                }
            }
        }
        if( !preq )
        {
            continue; /* No matching request */
        }

        /* Remove the pending entry */
        preq->cseq[j] = 0;

        /* Undo the damage we did on the way out */
        diff += replace_client_port( ppskb, ptran, tranlen, preq->sport_lo, preq->sport_hi );
        get_skb_data( ppskb, &pdata, &skblen );
        pdata += (skblen - datalen);
    }

    return diff;
}

/*
 * Handle outgoing (C->S) packet
 */
static int
masq_rtsp_out( struct ip_masq_app* papp, struct ip_masq* pms, struct sk_buff** ppskb, __u32 maddr )
{
    int   diff     = 0;
    __u32 skblen   = 0;
    char* pdata    = NULL;
    int   datalen  = 0;
    __u32 cseq     = 0;
    char* ptran    = NULL;
    __u32 tranlen  = 0;
    __u16 port_lo  = 0;     /* host endian */
    __u16 port_hi  = 0;     /* host endian */

    int i;
    struct rtsp_port_map* pmap;
    struct rtsp_port_map* preq;
    __u32 rtspf;

    get_skb_data( ppskb, &pdata, &skblen );
    datalen = skblen;

    while( datalen )
    {
        pmap = (struct rtsp_port_map*)pms->app_data;
        rtspf = parse_rtsp_message( &pdata, &datalen, "SETUP ", &cseq, &ptran, &tranlen );

        /* Is it a setup request with transport and cseq? */
        if( rtspf != RTSPF_ALL )
        {
            if( rtspf == 0 )
            {
                break; /* Parse failed, abort */
            }
            continue; /* Parse succeeded, keep going */
        }

        /* Does the transport have a client_port? */
        if( ! get_transport_ports( ptran, tranlen, &port_lo, &port_hi ) )
        {
            continue;
        }
        /* Is the client_port spec reasonable? */
        if( port_lo > port_hi || (port_hi-port_lo) > 1 )
        {
            continue;
        }

        /* Find map entry for this port range or grab a new one */
        preq = NULL;
        for( i = 0; i < MAX_PORT_MAPS; i++ )
        {
            if( pmap[i].sport_lo == 0 ||
                (pmap[i].sport_lo == port_lo && port_lo == port_hi) ||
                (pmap[i].sport_lo == port_lo && pmap[i].sport_hi == port_hi) )
            {
                preq = &pmap[i];
                break;
            }
        }
        if( ! preq )
        {
            /* Sorry, you used all your ports */
            printk( KERN_INFO "ip_masq_rtsp: masq_rtsp_out: no free client map entries\n" );
            continue;
        }

        /* Allocate new port(s) if needed */
        if( preq->sport_lo == 0 )
        {
            if( ! alloc_ports( maddr, pms, htons(port_lo), htons(port_hi), preq ) )
            {
                printk( KERN_INFO "ip_masq_rtsp: cannot allocate port(s)\n" );
                continue;
            }
        }

        /*
         * Everything is cool, so note the cseq and munge the packet.
         * It's not fatal if all the cseq slots are already full, but
         * the client will see the mport values in the response.
         */
        for( i = 0; i < MAX_SIMUL_SETUP; i++ )
        {
            if( preq->cseq[i] == 0 )
            {
                preq->cseq[i] = cseq;
                break;
            }
        }

        diff += replace_client_port( ppskb, ptran, tranlen, preq->mport_lo, preq->mport_hi );
        get_skb_data( ppskb, &pdata, &skblen );
        pdata += (skblen - datalen);
    }

    return diff;
}

/*** Module load/unload support ***/

static struct ip_masq_app g_masq_functbl =
{
    NULL,               /* next */
    "masq_rtsp_name",   /* name */
    0,                  /* type */
    0,                  /* n_attach */
    masq_rtsp_init_1,   /* ip_masq_init_1 */
    masq_rtsp_done_1,   /* ip_masq_done_1 */
    masq_rtsp_out,      /* pkt_out */
    masq_rtsp_in        /* pkt_in */
};

/*
 * Initialization (load) function
 */
static int __init
masq_rtsp_init( void )
{
    int i;

    memset( g_masq_objs, 0, sizeof(g_masq_objs) );
    for( i = 0; i < MAX_MASQ_APP_PORTS && ports[i]; i++ )
    {
        int s;
        g_masq_objs[i] = kmalloc( sizeof(struct ip_masq_app), GFP_KERNEL );
        if( ! g_masq_objs[i] )
        {
            return -ENOMEM;
        }
        memcpy( g_masq_objs[i], &g_masq_functbl, sizeof(struct ip_masq_app) );
        s = register_ip_masq_app( g_masq_objs[i], IPPROTO_TCP, ports[i] );
        if( s )
        {
            return s;
        }
    }

    return 0;
}

/*
 * Destruction (unload) function
 */
static int
masq_rtsp_done( void )
{
    int i;
    int rv;

    rv = 0;
    for( i = 0; i < MAX_MASQ_APP_PORTS; i++ )
    {
        if( g_masq_objs[i] )
        {
            int s = unregister_ip_masq_app( g_masq_objs[i] );
            if( s )
            {
                rv = s;
            }
            else
            {
                kfree( g_masq_objs[i] );
                g_masq_objs[i] = NULL;
            }
        }
    }

    return rv;
}

#ifdef MODULE

EXPORT_NO_SYMBOLS;

int init_module( void )
{
    if( masq_rtsp_init() != 0 )
    {
        return -EIO;
    }

    return 0;
}

void cleanup_module( void )
{
    if( masq_rtsp_done() != 0 )
    {
        printk( KERN_INFO "ip_masq_rtsp: can't remove module" );
    }
}

#endif /* MODULE */
